package entity

import (
	"testing"

	"rsprd.com/spread/pkg/deploy"
	"rsprd.com/spread/pkg/image"

	"github.com/stretchr/testify/assert"
	kube "k8s.io/kubernetes/pkg/api"
)

func TestImageDeployment(t *testing.T) {
	imageName := "arch"
	simple := newDockerImage(t, imageName)

	image, err := NewImage(simple, kube.ObjectMeta{}, "test")
	assert.NoError(t, err, "valid image")

	expectedPod := kube.Pod{
		ObjectMeta: kube.ObjectMeta{
			GenerateName: imageName,
			Namespace:    "default",
		},
		Spec: kube.PodSpec{
			Containers: []kube.Container{
				{
					Name:            imageName,
					Image:           imageName,
					ImagePullPolicy: kube.PullAlways,
				},
			},
			RestartPolicy: kube.RestartPolicyAlways,
			DNSPolicy:     kube.DNSDefault,
		},
	}

	expected := new(deploy.Deployment)
	assert.NoError(t, expected.Add(&expectedPod), "should be able to add pod")

	actual, err := image.Deployment()
	assert.NoError(t, err, "deploy ok")
	testDeploymentEqual(t, expected, actual)
}

func TestImageImages(t *testing.T) {
	imageName := "gcr.io/google_containers/cassandra:v7"
	simple := newDockerImage(t, imageName)

	image, err := NewImage(simple, kube.ObjectMeta{}, "test")
	if err != nil {
		t.Fatalf("Could not create Image entity: %v", err)
	}

	// check images
	images := image.Images()
	assert.Len(t, images, 1, "supposed to be single image")
	assert.EqualValues(t, simple, images[0], "should return image it was created with")
}

func TestImageNil(t *testing.T) {
	var image *image.Image
	_, err := NewImage(image, kube.ObjectMeta{}, "")
	assert.Error(t, err, "cannot be nil")
}

func TestImageInvalid(t *testing.T) {
	image := new(image.Image)
	_, err := NewImage(image, kube.ObjectMeta{}, "")
	assert.Error(t, err, "not valid")
}

func TestImageAttach(t *testing.T) {
	a := testNewImage(t, "a", kube.ObjectMeta{}, "", testRandomObjects(30))
	b := testNewImage(t, "b", kube.ObjectMeta{}, "", testRandomObjects(30))

	err := a.Attach(b)
	assert.Error(t, err, "Nothing can attach to images")
}

func TestImageType(t *testing.T) {
	image := newDockerImage(t, "ghost:latest")

	entity, err := NewImage(image, kube.ObjectMeta{}, "")
	if err != nil {
		t.Fatalf("Could not create Image entity: %v", err)
	}

	assert.Equal(t, EntityImage, entity.Type(), "incorrect type")
}

func TestImageKube(t *testing.T) {
	imageName := "redis:latest"
	image := newDockerImage(t, imageName)

	entity, err := NewImage(image, kube.ObjectMeta{}, "", createSecret("helo"))
	if err != nil {
		t.Fatalf("Could not create Image entity: %v", err)
	}

	actual, objects, err := entity.data()
	assert.NoError(t, err, "images can't have bad data")
	assert.Equal(t, objects.Len(), 1, "should only have secret")
	assert.Equal(t, imageName, actual, "image names should match")
}

func TestImageBadObject(t *testing.T) {
	// TODO: RESOLVE
	t.Skip("Currently deployment does not check for validity")

	imageName := "debian"
	image := newDockerImage(t, imageName)

	service := kube.Service{}

	_, err := NewImage(image, kube.ObjectMeta{}, "", &service)
	assert.Error(t, err, "invalid object, should return error")
}

func TestImageChildren(t *testing.T) {
	image := newDockerImage(t, "ghost:latest")

	entity, err := NewImage(image, kube.ObjectMeta{}, "")
	if err != nil {
		t.Fatalf("Could not create Image entity: %v", err)
	}

	children := entity.children()
	assert.Len(t, children, 0, "images can't have children")
}

func testNewImage(t *testing.T, imageName string, defaults kube.ObjectMeta, source string, objects []deploy.KubeObject) *Image {
	image, err := NewImage(newDockerImage(t, imageName), defaults, source, objects...)
	if err != nil {
		t.Fatalf("Could not create Image: %v", err)
	}

	return image
}

func newDockerImage(t *testing.T, imageName string) *image.Image {
	simple, err := image.FromString(imageName)
	if err != nil {
		t.Fatalf("Could not create image.Image: %v", err)
	}
	return simple
}
